@@ -41,20 +41,26 @@
 '''
 Known Issues:
     1) Multi-core compilation on Windows not supported
+    2) Won't work on Windows 2000 or windows where home directory
+       has spaces -- if we use ~ the build command will fail.
+
+Current Goals:
+    1) Get
 '''
 
 from optparse import OptionParser, OptionGroup #note: deprecated in Python27, use argparse
 from mozrunner import Runner, FirefoxRunner
 from mozrunner import FirefoxProfile
-from types import *
-from utils import hgId, captureStdout
-import os, sys, subprocess, string, re, tempfile, shlex, glob, shutil,datetime
 import simplejson, urllib
 import ximport
 
+from types import *
+from utils import hgId, captureStdout
+import os, sys, subprocess, string, re, tempfile, shlex, glob, shutil, datetime, multiprocessing
+
 #Global Variables
 showMakeData = 0
-progVersion="0.4.1"
+progVersion="0.4.4"
 
 class Builder():
     def __init__(self, makeCommand=["make","-f","client.mk","build"] , shellCacheDir=os.path.join(os.path.expanduser("~"), "moz-commitbuilder-cache"), cores=1, repoURL="http://hg.mozilla.org/mozilla-central",clean=False, mozconf=None):
@@ -69,20 +75,10 @@ class Builder():
         self.mozconf = mozconf
         self.objdir = os.path.join(self.repoPath,"obj-ff-dbg")
 
-        self.mochitest = os.path.join(self.objdir,"_tests","testing","mochitest","tests")
-        self.mochitest_tmp = os.path.join(self.objdir,"_tests","testing","mochitest","tests","commitbuilder")
-        #self.ui = ui.ui()
-        #self.hgrepo = hg.repository(self.ui, self.repoPath)
-        #To use hg-python, you do: commands.MYCMDNAME(self.ui, self.repo, {args})
-
         if not os.path.exists(shellCacheDir):
             os.mkdir(shellCacheDir)
         if not os.path.exists(self.confDir):
             os.mkdir(self.confDir)
-        #if not os.path.exists(self.mochitest_tmp):
-        #    os.mkdir(self.mochitest_tmp)
-            print self.mochitest_tmp
-            print "created!"
 
         #Sanity check: make sure hg is installed on the system, otherwise do not proceed!
         try:
@@ -120,7 +116,8 @@ class Builder():
         return str(nextDate)
 
     def changesetFromDay(self, date, oldest=True):
-        #Gets first changeset from a given date
+        # Gets first changeset from a given date
+        # NOTE: USES PUSHLOG
         nextDate = self.increment_day(date)
         pushlog_url = "http://hg.mozilla.org/mozilla-central/json-pushes?startdate="+date+"&enddate="+nextDate
         pushlog_json = simplejson.load(urllib.urlopen(pushlog_url))
@@ -175,15 +172,16 @@ class Builder():
             return
 
         #Make mozconfig
-        os.chdir(self.confDir)
-        if os.path.exists("config-default"):
-            os.unlink("config-default")
+        mozconfig_path = os.path.join(self.confDir, 'config-default')
+        if os.path.exists(mozconfig_path):
+            os.unlink(mozconfig_path)
 
-        f=open('config-default', 'w')
+        f=open(mozconfig_path, 'w')
         #Ensure we know where to find our built stuff by using a custom mozconfig
         f.write('mk_add_options MOZ_OBJDIR=@TOPSRCDIR@/obj-ff-dbg\n')
         f.write('ac_add_options --disable-optimize\n')
         f.write('ac_add_options --enable-debug\n')
+        f.write('ac_add_options --enable-tests\n')
 
         #HACK :/
         if sys.platform == "win32" or sys.platform == "cygwin":
@@ -197,81 +195,82 @@ class Builder():
         f.close()
 
         #export MOZCONFIG=/path/to/mozilla/mozconfig-firefox
-        os.environ['MOZCONFIG']=self.confDir+"/config-default"
+        os.environ['MOZCONFIG']=mozconfig_path
 
-    def bisect(self,good,bad, testfile=None, testpath=None, testcondition=None, args_for_condition=None):
+    def importAndBisect(self, good, bad, testcondition=None, args_for_condition=[]):
+        #convenience function for API use only
+        conditionscript = ximport.importRelativeOrAbsolute(testcondition)
+        self.bisect(good,bad, testcondition=conditionscript, args_for_condition=args_for_condition)
+
+
+
+    def bisect(self,good,bad, testcondition=None, args_for_condition=[]):
         #Call hg bisect with initial params, set up building environment (mozconfig)
         #testfile is an external file, testpath is an actual mochitest
+
+        #Support for using dates
+        #EXPERIMENTAL
+        badDate = re.search(r'(\d\d\d\d-\d\d-\d\d)',good)
+        goodDate = re.search(r'(\d\d\d\d-\d\d-\d\d)',bad)
+        if badDate != None and goodDate != None:
+            badDate = badDate.group(1)
+            goodDate = goodDate.group(1)
+            good = self.changesetFromDay(goodDate) #Get first changeset from this day
+            bad = self.changesetFromDay(badDate, oldest=False) #Get last changeset from that day
+
         good = hgId(good, self.hgPrefix)
         bad = hgId(bad, self.hgPrefix)
         if good and bad and self.validate(good, bad): #valid commit numbers, do the bisection!
             subprocess.call(self.hgPrefix+["bisect","--reset"])
-            subprocess.call(self.hgPrefix+["up",bad])
-            subprocess.call(self.hgPrefix+["bisect","--bad"])
-            subprocess.call(self.hgPrefix+["bisect","--good",good])
 
-            #Prebuild stuff here!!
+            setUpdate = captureStdout(self.hgPrefix+["up",bad])
+            setBad = captureStdout(self.hgPrefix+["bisect","--bad"])
+            setGood = captureStdout(self.hgPrefix+["bisect","--good",good])
 
+            print str(setUpdate)
+            print str(setBad)
+            print str(setGood)
+
+            # Check if we should terminate early because the bisector exited?
+            string_to_parse = str(setGood)
+            traceback_flag = string_to_parse.find("Not all ancestors")
+            if traceback_flag > -1:
+                #TODO: Bisect from ancestor changeset rather than exiting early
+                quit()
+            traceback_flag = string_to_parse.find("The first bad revision is")
+            if traceback_flag > -1:
+                quit()
+
+            # Set mozconfig
+            # Call recursive bisection!
             self.mozconfigure()
-            self.bisectRecurse(testfile=testfile, testpath=testpath)
+            self.bisectRecurse(testcondition=testcondition, args_for_condition=args_for_condition)
+
         else:
             print "Invalid values. Please check your changeset revision numbers."
 
-    def bisectRecurse(self, testfile=None, testpath=None, testcondition=None, args_for_condition=[]):
+    def bisectRecurse(self, testcondition=None, args_for_condition=[]):
         #Recursively build, run, and prompt
         verdict = ""
 
-        if testfile==None and testpath==None and testcondition==None:
+        if testcondition==None:
             #Not using a test, interactive bisect begin!
             self.buildAndRun()
         elif testcondition != None:
             #Using Jesse's idea: import any testing script and run it as the truth condition
-            conditionscript = ximport.importRelativeOrAbsolute(testcondition)
-            verdict = conditionscript.interesting([self.objdir] + args_for_condition)
-        else:
-            #TODO UNCOMMENT LINE BELOW
             self.build()
-            if testfile == None:
-                #Using External testfile
-                #1. Clear self.mochitest_tmp
-                try:
-                    shutil.rmtree(self.mochitest_tmp)
-                except:
-                    pass
-                #2. Move file from testpath to self.mochitest_tmp
-                try:
-                    os.mkdir(self.mochitest_tmp)
-                    dst =  os.path.join(self.mochitest_tmp,"test_bug999999.html")
-                    subprocess.call(['touch',dst])
-                    print "copy " + str(testpath) + " to " +dst
-                    shutil.copy(str(testpath),dst)
-                except:
-                    print "Unable to generate test path, quitting. Is your inputted test a valid mochitest?"
-                    quit()
-                #3. Make mochitest use the commitbuilder directory.
-                testfile="commitbuilder"
+            args_to_pass = [self.objdir] + args_for_condition
 
-            #DEBUG
-            #testfile = "content/base/test/test_CrossSiteXHR.html"
+            if hasattr(testcondition, "init"):
+                testcondition.init(args_to_pass)
 
-            print "Trying testfile "+str(testfile)
-            #TODO:
-            # 1. Copy relevant test to a directory of my choice
+            #TODO: refactor to use directories with revision numbers
+            tmpdir = tempfile.mkdtemp()
+            verdict = testcondition.interesting(args_to_pass,tmpdir)
 
-
-            #sts = os.system("TEST_PATH="+testpath+" EXTRA_TEST_ARGS='--close-when-done' make -C " +self.objdir+ " mochitest-plain")
-            sts = subprocess.call(['make','-C',self.objdir,'mochitest-plain'],stdout=open('/dev/null','w'),env={'TEST_PATH': testfile, 'EXTRA_TEST_ARGS':"--close-when-done"})
-            shutil.rmtree(self.mochitest_tmp) #cleanup
-            if sts != 0:
-                verdict = "bad"
-                print "============================"
-                print "Verdict: FAILED test, bad changeset detected!"
-                print "============================"
-            else:
-                verdict = "good"
-                print "============================"
-                print "Verdict: PASSED test, good changeset detected!"
-                print "============================"
+            #Allow user to return true/false or bad/good
+            if verdict != "bad" and verdict != "good":
+                verdict = "bad" if verdict else "good"
 
         if verdict != 'good' and verdict !='bad':
             while verdict != 'good' and verdict != 'bad' and verdict != 'b' and verdict != 'g':
@@ -285,6 +284,10 @@ class Builder():
             retval = captureStdout(self.hgPrefix+["bisect","--bad"])
 
         print str(retval)
+        string_to_parse = str(retval)
+        traceback_flag = string_to_parse.find("Not all ancestors")
+        if traceback_flag > -1:
+            print "You need to re-run the bisector again using the changeset they give you."
 
         # HACK
         if retval[1] == 'h': #if retval starts with "the" then we can quit
@@ -295,7 +298,7 @@ class Builder():
             print "Something went wrong! :("
             quit()
 
-        self.bisectRecurse(testfile=testfile, testpath=testpath, testcondition=testcondition, args_for_condition=args_for_condition)
+        self.bisectRecurse(testcondition=testcondition, args_for_condition=args_for_condition)
 
     def buildAndRun(self, changeset=0):
         #API convenience function
@@ -324,7 +327,7 @@ class Builder():
     def run(self):
         #Run the built binary if it exists. ONLY WORKS IF BUILD WAS CALLED!
         if sys.platform == "darwin":
-            runner = FirefoxRunner(binary=os.path.join(self.shellCacheDir,"mozbuild-trunk","obj-ff-dbg","dist","Nightly.app","Contents","MacOS")+"/firefox-bin")
+            runner = FirefoxRunner(binary=os.path.join(self.shellCacheDir,"mozbuild-trunk","obj-ff-dbg","dist","NightlyDebug.app","Contents","MacOS")+"/firefox-bin")
             runner.start()
             runner.wait()
         elif sys.platform == "linux2":
@@ -398,6 +401,12 @@ class Builder():
             return False
         return True
 
+def cpuCount():
+    try:
+        return multiprocessing.cpu_count()
+    except NotImplementedError:
+        return 1
+
 def cli():
     #Command line interface
     usage = """usage: %prog --good=[changeset] --bad=[changeset] [options] \n       %prog --single=[changeset] -e
@@ -408,20 +417,21 @@ def cli():
 
     group1 = OptionGroup(parser, "Global Options",
                                         "")
-    group1.add_option("-j", "--cores", dest="cores",
-                                        help="Number of cores to compile with",
-                                        metavar="[numcores]")
+    group1.add_option("-j", "--cores", dest="cores", default=cpuCount(),
+                                        help="Max simultaneous make jobs (default: %default, the number of cores on this system)",
+                                        metavar="[numjobs]")
     group1.add_option("-f", "--freshtrunk", action = "store_true", dest="makeClean", default=False,
                                         help="Delete old trunk and use a fresh one")
 
     group2 = OptionGroup(parser, "Bisector Options",
-                                        "These are options for bisecting on changesets")
+                                        "These are options for bisecting on changesets. Dates are retrieved from pushlog " \
+                                        "and are not as reliable as exact changeset numbers.")
     group2.add_option("-g", "--good", dest="good",
                                         help="Last known good revision",
-                                        metavar="[changeset]")
+                                        metavar="[changeset or date]")
     group2.add_option("-b", "--bad", dest="bad",
                                         help="Broken commit revision",
-                                        metavar="[changeset]")
+                                        metavar="[changeset or date]")
 
     group3 = OptionGroup(parser, "Single Changeset Options",
                                         "These are options for building a single changeset")
@@ -443,30 +453,24 @@ def cli():
                                         help="revision number for single changeset to build binary from")
 
     group5 = OptionGroup(parser, "Automatic Testing Options (EXPERIMENTAL)",
-                                        "Options for using an automated test instead of interactive prompting for bisection")
+                                        "Options for using an automated test instead of interactive prompting for bisection. " \
+                                        "Please read documentation on how to write testing functions for this script.")
 
-    group5.add_option("-t", "--testfile", dest="testfile", default=None, metavar="content/base/test",
-                                        help="relative path of test directory in mochitest")
 
-    group5.add_option("-w", "--exttest", dest="testpath", default=None, metavar="~/Desktop/mytest.html",
-                                        help="absolute path to your own test file")
+    group5.add_option("-c", "--condition", dest="condition", default=None, metavar="[condtest.py -opt1 -opt2]",
+                                        help="External condition for bisecting. " \
+                                             "Note: THIS MUST BE THE LAST OPTION CALLED.")
 
-    group5.add_option("-c", "--condition", dest="condition", default=None, metavar="[cond.py -opt1 -opt2]",
-                                        help="external condition for bisecting: MAKE THIS LAST OPTION")
-
-    group6 = OptionGroup(parser, "Unstable Options",
+    group6 = OptionGroup(parser, "Broken and Unstable Options",
                                         "Caution: use these options at your own risk.  "
                                         "They aren't recommended.")
 
     group6.add_option("-r", "--repo", dest="repoURL",
-                                        help="alternative mercurial repo to bisect",
+                                        help="alternative mercurial repo to bisect NOTE: NEVER BEEN TESTED",
                                         metavar="valid repository url")
 
-    #group.add_option("-m", "--altmake", dest="alternateMake",
-    #                                    help="alternative make command for building",
-    #                                    metavar="make command, in quotes")
     group6.add_option("-m", "--mozconfig", dest="mozconf",
-                                        help="external mozconfig if so desired",
+                                        help="external mozconfig if so desired NOTE: BROKEN RIGHT NOW",
                                         metavar="path_to_mozconfig", default=False)
 
     parser.add_option_group(group1)
@@ -480,9 +484,11 @@ def cli():
     # If a user only wants to make clean or has supplied no options:
     if (not options.good or not options.bad) and not options.single:
         if options.makeClean:
-            #Make a clean trunk
+            #Make a clean trunk and quit.
             commitBuilder = Builder(clean=options.makeClean)
         else:
+            # Not enough relevant parameters were given
+            # Print a help message and quit.
             print """Use -h flag for available options."""
             print """To bisect, you must specify both a good and a bad date. (-g and -b flags are not optional)."""
             print """You can also use the --single=[chset] flag to build and --run to run a single changeset."""
@@ -522,7 +528,10 @@ def cli():
         #if options.alternateMake:
         #    commitBuilder.makeCommand = shlex.split(options.alternateMake)
 
-        commitBuilder.bisect(options.good,options.bad, testfile=options.testfile, testpath=options.testpath, testcondition=options.condition, args_for_condition=args_for_condition)
+        if options.condition:
+            conditionscript = ximport.importRelativeOrAbsolute(options.condition)
+
+        commitBuilder.bisect(options.good,options.bad, testcondition=conditionscript, args_for_condition=args_for_condition)
 
     # Should not get here.
     else:
